% Copyright (C) 1998, Mario A. Storti
%
% This file is part of Octave.
%
% Octave is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public
% License as published by the Free Software Foundation;
% either version 2, or (at your option) any later version.
%
% Octave is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied
% warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public
% License along with Octave; see the file COPYING.  If not,
% write to the Free Software Foundation, 59 Temple Place -
% Suite 330, Boston, MA 02111-1307, USA.
%
%usage: [x,rhist]= cgne (A,b,x0,toler)
%
%    Solve the linear system A*x=b iteratively by the method of Conjugate 
%    Gradient on Normal Equations. 
%
%    A (matrix/string)      : The matrix of coefficients of the linear
%                                 system. 
%                                 If it is a string, say 'atimesx' 
%                                 then it should be
%                                 the name of an octave function returning A*x
%                                 when passed x as an argument.
%                                 atimesx(x) should return A*x, and 
%                                 atimes(x,'t') should return A'*x.
%    b (vector/matrix)      : right hand side. If it is a matrix, then
%                             it represents multiple right hand sides,
%                             and x returns a matrix containing the
%                             corresponding solutions. 
%    x0 (vector, optional)  : initial guess for x
%    tolerance ([scalar scalar])
%                           : tolerance on the residual 
%                              (default: 1e-10), maximum number of 
%                              iterations (default: infinity). 
%    x (vector)             : solution of the linear system 
%                                 to prescribed tolerance
%    rhist (vector)         : history of convergence. rhist(k) returns the
%                                 residual at iteration k.

% Author: Mario Storti <mstorti@galileo.unl.edu.ar> 
% Keywords: linear systems, iterative methods
% Created: Sun May 31 19:54:39 ART 1998

function [x,rhist]= cgne (rname_,b,x0,tolerance)

%$Id: cgne.m,v 1.4 1998/06/01 16:12:41 mstorti Exp mstorti $

  if nargin<=1 | nargin >4
    usage ('[x,rhist]= cgne (''atimesx'',b,x0,tolerance)');
    return
  end 

  if nargin<=3
    tolerance=[1e-10 -1];
  end

  if length(tolerance)==1
    tolerance(2)=-1;
  end 

  if all(tolerance<0)
    error('one of the components of tolerance must be greater than zero')
  end 

  if nargin<=2
    x0=zeros(size(b));
  end 

  if any(size(x0)~=size(b))
    error('x0 and b should be of the same size')
  end 

  % Initialization
  s=x0;
  beta=sqrt(sum(b.*b));
  if isstr(rname_)
    t1=feval(rname_,s);
    t2=b-t1;
    r1=feval(rname_,t2,'t');
  else
    t1=rname_*s;
    t2=b-t1;
    r1=rname_'*t2;
  end 
%  r2=feval(preco_,r1);
  r2=r1;
  r3=-r2;
  D0=sum(r1.*r2);
  Ogb=sqrt(D0)/beta;

  % Main loop
  iter=0;
  rhist=[];
  if (max(Ogb)<tolerance(1) & tolerance(1)>0) 
    x=x0;
    rhist=Ogb;
    return
  end 

  while 1
    iter=iter+1;
    if isstr(rname_)
      t1=feval(rname_,r3);
      t2=feval(rname_,t1,'t');
    else
      t1=rname_*r3;
      t2=rname_'*t1;
    end
    sigma=sum(r3.*t2);
    lambda=D0./sigma;
    s=s+rightscal(r3,lambda');
    r1=r1+rightscal(t2,lambda');
%    t2=feval(preco_,r1);
    t2=r1;
    D1=sum(r1.*t2);
    % Energy norm of the residual
    Ogb=sqrt(D1)./beta;
    % Save convergence history
    rhist=[rhist;Ogb];
    % Check convergence
    if (max(Ogb)<tolerance(1) & tolerance(1)>0) | ...
          (iter>tolerance(2) & tolerance(2)>0)
      break;
    end
    alpha=D1./D0;
    D0=D1;
    r3=rightscal(r3,alpha')-t2;
  end
  x=-s;
%end
